local CleanMyBag = LibStub("AceAddon-3.0"):NewAddon("Clean My Bag", "AceConsole-3.0", "AceEvent-3.0", "AceConsole-3.0");
local L = LibStub('AceLocale-3.0'):GetLocale('Clean My Bag', true)
local AceConfig = LibStub("AceConfig-3.0")
local AceGUI = LibStub("AceGUI-3.0")

local options = {
	name = "Clean My Bag",
	handler = CleanMyBag,
	type = "group",
	args = {
		c = {
			name = "Settings",
			type = "group",
			inline = true,
			args = {
				fenabled = {
					name = 'Trashing Enabled',
					desc = "Enable CleanMyBag filtering on loot items.",
					type = 'toggle',
					set = 'SetOption',
					get = 'GetOption',
					width = "full",
					order = 1,
				},
				debug_mode = {
					name = 'Debug Mode',
					desc = "Loot will not be trashed, you'll receive notices.",
					type = 'toggle',
					set = 'SetOption',
					get = 'GetOption',
					width = "full",
					order = 2,
				},
				silent_mode = {
					name = 'Silent Mode',
					desc = "Trashing chat notices.",
					type = 'toggle',
					set = 'SetOption',
					get = 'GetOption',
					width = "full",
					order = 3,
				},
				enhanced_tooltip = {
					name = 'Enhanced Loot Tooltip',
					desc = 'Displays additional info within the loot tooltip.',
					type = 'toggle',
					set = 'SetOption',
					get = 'GetOption',
					width = 'full',
					order = 4,
				},
				filter_bags = {
					name = "Bags",
					type = "group",
					order = 5,
					args = {
						bag0 = {
							name = 'Bag 1',
							type = 'toggle',
							set = 'SetOption',
							get = 'GetOption',
							width = "half",
							order = 1,
						},
						bag1 = {
							name = 'Bag 2',
							type = 'toggle',
							set = 'SetOption',
							get = 'GetOption',
							width = "half",
							order = 2,
						},
						bag2 = {
							name = 'Bag 3',
							type = 'toggle',
							set = 'SetOption',
							get = 'GetOption',
							width = "half",
							order = 3,
						},
						bag3 = {
							name = 'Bag 4',
							type = 'toggle',
							set = 'SetOption',
							get = 'GetOption',
							width = "half",
							order = 4,
						},
						bag4 = {
							name = 'Bag 5',
							type = 'toggle',
							set = 'SetOption',
							get = 'GetOption',
							width = "half",
							order = 5,
						},
						bag_all = {
							name = 'All',
							type = 'toggle',
							set = 'SetOption',
							get = 'GetOption',
							width = "half",
							order = 6,
						},
					},
				},
				cfilters = {
					name = 'Current Filters - Malformed filters are automatically removed',
					type = 'input',
					multiline = 14,
					width = 'full',
					desc = 'add, remove and edit filters, one filter per line.',
					set = 'SetFilters',
					get = 'GetFilters',
					order = 6,
				},
				-- cfiltersLabel = {
				-- 	name = 'Example:\n\n|cffFF4D00type = "Trade Goods" and subtype = "Cooking"\n\ntype = "Miscellaneous" and subtype = "Junk" and value < "1000"\n\ntype = "Miscellaneous" and subtype = "Reagent" and value < "3000"\n\n|r',
				-- 	type = 'description',
				-- 	width = 'full',
				-- 	order = 7,
				-- },
				help_label = {
					name = L["TagHelp"],
					type = 'description',
					order = 7,
					width = 'full',
				},
				safelist = {
					name = 'Safe List - Items in this list will never be removed',
					type = 'input',
					multiline = 10,
					width = 'full',
					desc = 'comma delimit item names.',
					set = 'SetSafeList',
					get = 'GetSafeList',
					order = 8,
				},
			}
		}
	}
}

local defaults = { 
	char = {
		fenabled=true,
		debug_mode=true,
		silent_mode=false,
		enhanced_tooltip=true,
		bag0=true,
		bag1=true,
		bag2=true,
		bag3=true,
		bag4=true,
		bag_all=true,
		safelist="",
		cfilters="",
		soulboundItems={},
		conjuredItems={},
		professionItems={},
		dev_debug=false,
	},
}
local loot, groupLoot, bags, filters, trash, safe = {}, {}, {}, {}, {}, {}
local soulboundItems, conjuredItems, blacksmithingItems = {}, {}, {}
local lootPushed, totalLoot, bagUpdates = 0, 0, 0
local lineAdded = false

local function OnTooltipSetItem(tooltip, whichbag, whichslot)
	if CleanMyBag.db.char.enhanced_tooltip then
		name, link = GameTooltip:GetItem()
		if link ~= nil then
			itemName, itemLink, itemRarity, itemLevel, itemMinLevel, itemType, itemSubType,
			itemStackCount, itemEquipLoc, itemTexture, itemSellPrice = GetItemInfo(link)
			local copper = itemSellPrice
			if not lineAdded then
				if itemType ~= nil and itemSubType ~= nil and itemSellPrice ~= nil then
					tooltip:AddLine(" ")
					tooltip:AddLine("Clean My Bag", 0.33, 0.74, 0.92)
					tooltip:AddLine("  Type: " .. itemType, 1, 1, 1)
					tooltip:AddLine("  Subtype: " .. itemSubType, 1, 1, 1)
					tooltip:AddLine("  Value: " .. itemSellPrice .. " copper", 1, 1 ,1)
					-- VALUE IN gold silver copper
					-- tooltip:AddLine("  Value: " .. ("%dg %ds %dc"):format(copper / 100 / 100, (copper / 100) % 100, copper % 100), 0.8, 0.8, 0.8)
					lineAdded = true
				end
			end
		end
	end
end

local function OnTooltipCleared(tooltip, ...)
	lineAdded = false
end

GameTooltip:HookScript("OnTooltipSetItem", OnTooltipSetItem)
GameTooltip:HookScript("OnTooltipCleared", OnTooltipCleared)

function CleanMyBag:OnInitialize()
	CleanMyBag.db = LibStub("AceDB-3.0"):New("CMBFiltersDB", defaults)
	CleanMyBag.version = GetAddOnMetadata("CleanMyBag", "Version")
	LibStub("AceConfig-3.0"):RegisterOptionsTable("CleanMyBag", options)
	CleanMyBag.optionsDialog = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("CleanMyBag")
	CleanMyBag:SetFilters('', CleanMyBag.db.char.cfilters)
	CleanMyBag:SetSafeList('', CleanMyBag.db.char.safelist)

	-- get_professions()
end

function CleanMyBag:OnEnable()
	-- print_str(": Version " .. CleanMyBag.version)
	CleanMyBag:RegisterEvent("BAG_UPDATE")
	CleanMyBag:RegisterEvent("LOOT_OPENED")
	CleanMyBag:RegisterEvent("ITEM_PUSH")
	CleanMyBag:RegisterEvent("START_LOOT_ROLL")
	CleanMyBag:RegisterEvent("CHAT_MSG_LOOT")
	
	CleanMyBag:RegisterChatCommand("cleanmybag", "OnSlashCommand")
	CleanMyBag:RegisterChatCommand("cmb", "OnSlashCommand")
	foodList = foodList or {}
	waterList = waterList or {}
end

function CleanMyBag:OnDisable()
	CleanMyBag:UnregisterEvent("BAG_UPDATE")
	CleanMyBag:UnregisterEvent("LOOT_OPENED")
	CleanMyBag:UnregisterEvent("ITEM_PUSH")
	CleanMyBag:UnregisterEvent("START_LOOT_ROLL")
	CleanMyBag:UnregisterEvent("CHAT_MSG_LOOT")
	
	CleanMyBag:UnregisterChatCommand("cleanmybag")
	CleanMyBag:UnregisterChatCommand("cmb")
end

function CleanMyBag:OnSlashCommand(input)
	local command, rest = input:match("^(%S*)%s*(.-)$");

	if command == "enable" then
		if CleanMyBag.db.char.fenabled == false then
			CleanMyBag.db.char.fenabled = true
			self:RegisterEvent("BAG_UPDATE")
			self:RegisterEvent("LOOT_OPENED")
			self:RegisterEvent("ITEM_PUSH")
			self:RegisterEvent("START_LOOT_ROLL")
			self:RegisterEvent("CHAT_MSG_LOOT")
			
			print_str(': enabled.')
		else
			print_str(': is already enabled.')
		end
	elseif command == "disable" then
		if CleanMyBag.db.char.fenabled == true then
			CleanMyBag.db.char.fenabled = false
			self:UnregisterEvent("BAG_UPDATE")
			self:UnregisterEvent("LOOT_OPENED")
			self:UnregisterEvent("ITEM_PUSH")
			self:UnregisterEvent("START_LOOT_ROLL")
			self:UnregisterEvent("CHAT_MSG_LOOT")
			
			print_str(': disabled.')
		else
			print_str(': is already disabled.')
		end
	elseif command == "config" then
		InterfaceOptionsFrame_OpenToCategory(CleanMyBag.optionsDialog)
	elseif command == "silent" then
		CleanMyBag.db.char.silent_mode = true
	elseif command == "test" then
		ProcessRollResult('TEST', "Voshie receives loot: [Chaos Orb].")
	else
		print_str(" Syntax: /cmb config, /cmb enable, /cmb disable");
	end
end

function CleanMyBag:LOOT_OPENED(event, auto)
	loot, bags = {}, {}
	totalLoot, lootPushed, bagUpdates = 0, 0, 0
	if CleanMyBag.db.char.fenabled and table.getn(filters) > 0 then
		-- print_str('GetNumLootItems(): ' .. GetNumLootItems())
		local lootIcon, lootName, lootQuantity, rarity, locked = GetLootSlotInfo(1)
		if lootQuantity ~= nil then
			if lootQuantity == 0 then
				totalLoot = GetNumLootItems() - 1
			else
				totalLoot = GetNumLootItems()
			end
		end
		for i=1,GetNumLootItems() do
			local lootIcon, lootName, lootQuantity = GetLootSlotInfo(i)
			if in_table(groupLoot, lootName) == false then
				if lootQuantity ~= nil then
					if lootQuantity > 0 then
						table.insert(loot, {lootName, false, false})
					end
				end
			else
				totalLoot = totalLoot - 1
			end
		end
		groupLoot = {}
	end
end

function CleanMyBag:BAG_UPDATE(event, bag)
	if CleanMyBag.db.char.fenabled and table.getn(filters) > 0 then
		bagUpdates = bagUpdates + 1
		if table.getn(loot) > 0 then
			if in_table(bags, bag) ~= true then
				table.insert(bags, bag)
			end
			if lootPushed >= totalLoot then
				filter_loot()
				lootPushed, totalLoot, bagUpdates = 0, 0, 0
			end
		end
	end
end

function CleanMyBag:ITEM_PUSH(self, event, ...)
	if table.getn(loot) > 0 then
		lootPushed = lootPushed + 1
	end
end

function CleanMyBag:START_LOOT_ROLL(event, rollID, rollTime)
	local link = GetLootRollItemLink(rollID)
	local lootName = GetItemInfo(link)
	table.insert(groupLoot, lootName)
end

function get_professions()

	local prof1, prof2, archaeology, fishing, cooking, firstAid = GetProfessions();

	local name, icon, skillLevel, maxSkillLevel, numAbilities, spelloffset, skillLine, skillModifier, specializationIndex, specializationOffset = GetProfessionInfo( prof1 );

	-- print_str(': ' .. name .. ' - first profession.')

end

function filter_loot()
	if CleanMyBag.db.char.fenabled and table.getn(filters) > 0 then
		local name, link, total_slots
		local llen = table.getn(loot)
		for i=1,table.getn(loot) do
			for bag = 0, 4 do
				if in_table(bags, bag) then
					if loot[i][2] ~= true then
						total_slots = GetContainerNumSlots(bag)
						for slot = 1, total_slots do
							link = GetContainerItemLink(bag, slot)
							if link ~= nil then
								name = GetItemInfo(link)
								if name == loot[i][1] then
									onItem(bag, slot, link)
									loot[i][2] = true
									break
								end
							end
						end
					end
				end
			end
		end
		loot, bags = {}, {}
	end
end

function special_item(bag, slot)	-- returns boolean
	local tooltip = CreateFrame( "GameTooltip", "ScanTooltip", nil, "GameTooltipTemplate" ); -- Tooltip name cannot be nil
	tooltip:SetOwner( WorldFrame, "ANCHOR_NONE" );

	-- Allow tooltip SetX() methods to dynamically add new lines based on these
	-- tooltip:AddFontStrings(
	-- 	tooltip:CreateFontString( "$parentTextLeft1", nil, "GameTooltipText" ),
	-- 	tooltip:CreateFontString( "$parentTextRight1", nil, "GameTooltipText" )
	-- );

	tooltip:ClearLines()
	tooltip:SetBagItem(bag, slot)

	local specialItem = EnumerateTooltipLines_helper( tooltip )

	tooltip:Hide()
	
	return specialItem
end

function EnumerateTooltipLines_helper( tooltip )
	local name = tooltip:GetName()
	local isSoulbound = false
	local isConjured = false

	-- Loop over each line in the tooltip:
	for i = 1, tooltip:NumLines() do
		-- Get a reference to the left-aligned text on this line:
		local left = _G[name .. "TextLeft" .. i]
		-- Print out the line number, the side (left/right), and the text that
		-- appears there, or "no text" if there is no text there:
		-- print(i, "LEFT", left:GetText() or "no text")
		if left ~= nil then
			isSoulbound = (left:GetText() == ITEM_SOULBOUND)
			isConjured = (left:GetText() == ITEM_CONJURED)
			if isSoulbound or isConjured then
				return { ["soulbound"] = isSoulbound, ["conjured"] = isConjured }
			end
		end
		-- Do the same for the right-aligned text:
		-- local right = _G[name .. "TextRight" .. i]
		-- print(i, "RIGHT", right:GetText() or "no text")
	end
	return false
end

function onItem(bag, slot, link)
	local itemName, itemLink, itemRarity, itemLevel, itemMinLevel, itemType, itemSubType, itemStackCount, itemEquipLoc, itemTexture, itemSellPrice = GetItemInfo(link)
	local isQuestItem, questId, isActive = GetContainerItemQuestInfo(bag, slot)
	local stackValue = itemStackCount * itemSellPrice
	local itemColor = string.match(itemLink, "|?c?f?f?(%x*)|")
	local itemId = GetContainerItemID(bag, slot);
	if CleanMyBag.db.char.dev_debug then
		-- print_str(': item: (' .. itemName .. ') type: (' .. itemType .. ') subtype: (' .. itemSubType .. ') id: (' .. itemId .. ') sell price: (' .. itemSellPrice .. ')')
	end

	local specialItem = special_item(bag, slot)

	if specialItem ~= false then
		if CleanMyBag.db.char.conjuredItems[itemId] == nil then
			if specialItem["conjured"] then
				CleanMyBag.db.char.conjuredItems[itemId] = true
				return false
			end
		else
			return false
		end
	end

	if specialItem ~= false then
		if CleanMyBag.db.char.soulboundItems[itemId] == nil then
			if specialItem["soulbound"] then
				-- we will not trash anything that is soulbound
				CleanMyBag.db.char.soulboundItems[itemId] = true
				return false
			end
		else
			return false
		end
	end

	if table.getn(filters) > 0 then
		if CleanMyBag.db.char['bag'..bag] then
			if questId ~= nil or isQuestItem ~= nil or itemType == "Quest" then
				-- print_str(': ' .. itemLink .. ' is a quest item')
				return
			end
			if isQuestItem == false or isQuestItem == nil or itemType ~= "Quest" then
				if itemRarity < 3  then
					-- and in_table(safe, itemName) == false then
					for index,safeName in ipairs(safe) do
						local safeItemName = trim_str( safeName, ' ' )
						if string.match(safeItemName, '..') then
							if string.match(itemName, safeItemName) then
								print_str(': ' .. itemLink .. ' met safe pattern')
								return false
							end
						else
							if itemName == safeName then
								return false
							end
						end
					end

					for mi,mf in pairs(filters) do
						local flen = table.getn(mf)
						local m = 0
							for index,filter in pairs(mf) do
								if filter[1] == "name" then
									-- if string.match(filter[3], '..') then
										-- if string.match(itemName, filter[3]) then
											-- m = m + 1
										-- end
									-- else
										if filter[3] == itemName then
											m = m + 1
										end
									-- end
								elseif filter[1] == "type" then
									if CleanMyBag.db.char.dev_debug then
										print_str(' type: ' .. itemType .. ' looking for: ' .. filter[3])
									end
									if filter[2] == "=" then
										if filter[3] == "Miscellaneous" then
											if itemRarity < 2 then
												m = m + 1
											end
											-- if CleanMyBag.db.char.dev_debug then
											-- 	print_str('itemColor: ' .. itemColor)
											-- end
											-- if ( itemType == "Miscellaneous" ) then
											-- 	print_str(': filter type = misc, itemColor: ' .. itemColor .. ' - item type: ' .. itemType)
											-- end
											-- if itemColor == "9d9d9d" then
											-- 	m = m + 1
											-- 	if CleanMyBag.db.char.dev_debug then
											-- 		print_str('itemColor match ' .. m)
											-- 	end
											-- end
										elseif filter[3] == "food" or filter[3] == "Food" and itemType == "Consumable" then
											if in_table(foodList, itemId) then
												m = m + 1
											end
										elseif filter[3] == "water" or filter[3] == "Water" and itemType == "Consumable" then
											if in_table(waterList, itemId) then
												m = m + 1
											end
										elseif filter[3] == itemType then
											m = m + 1
										end
									elseif filter[2] == "!=" and filter[3] ~= itemType then
										m = m + 1
									end
								elseif filter[1] == "subtype" then
									if filter[2] == "=" and filter[3] == itemSubType then
										if CleanMyBag.db.char.dev_debug then
											print_str(' ' .. link .. ' subtype found: ' .. itemSubType .. ' looking for subtype: ' .. filter[3])
										end
										m = m + 1
										if CleanMyBag.db.char.dev_debug then
											print_str('subtype match ' .. m)
										end
									elseif filter[2] == "!=" and filter[3] ~= itemSubType then
										m = m + 1
									end
								elseif filter[1] == "quality" then
									if CleanMyBag.db.char.dev_debug then
										print_str('quality found: ' .. itemRarity .. ' looking for quality: ' .. filter[3])
									end
									if filter[2] == "=" and tonumber(filter[3]) == itemRarity then
										m = m + 1
										if CleanMyBag.db.char.dev_debug then
											print_str('quality match ' .. m)
										end
									elseif filter[2] == "!=" and tonumber(filter[3]) ~= itemRarity then
										m = m + 1
									end
								elseif filter[1] == "value" then
									if itemSellPrice < tonumber(filter[3]) then
										m = m + 1
									end
								elseif filter[1] == "level" and itemLevel < tonumber(filter[3]) then
									m = m + 1
								end
							end
							if CleanMyBag.db.char.dev_debug then
								print_str(': ' .. link .. ' needed (' .. flen .. ') found (' .. m .. ')')
							end
							if m == flen then
								if CleanMyBag.db.char.debug_mode then
									print_str(': ' .. link .. ' marked to destroy, no action taken - debug active.')
								else
									DeleteContainerItem(bag, slot, link)
									return true
								end
							end
					end
				else
					-- if itemType == "Quest" then
					-- 	print_str( ' ' .. link .. ' itemType = Quest' )
					-- end

					-- if isQuestItem ~= false then
					-- 	print_str( ' ' .. link .. ' isQuestItem: ' .. isQuestItem )
					-- end

					-- if itemRarity >= 4 then
					-- 	print_str( link .. ' is a rare: ' .. itemRarity )
					-- end
					
					-- if in_table(safe, itemName) ~= false then
					-- 	print_str( link .. ' is in safe table' )
					-- end
				end -- item isn't rare and not in the safe table
			end -- item isn't a quest item
		end
	end
end

function DeleteContainerItem(bag, slot, link, filter)
	PickupContainerItem(bag, slot);
	if CursorHasItem() then
		DeleteCursorItem();
	end
	local itemName, itemLink, itemRarity, itemLevel, itemMinLevel, itemType, itemSubType, itemStackCount, itemEquipLoc, itemTexture, itemSellPrice = GetItemInfo(link);
	if CleanMyBag.db.char.silent_mode ~= true then
		print_str(': ' .. link .. ' destroyed.')
	end
end

function CleanMyBag:GetOption(info)
	return CleanMyBag.db.char[info[#info]]
end

function CleanMyBag:SetOption(info, input)
	CleanMyBag.db.char[info[#info]] = input
	if info[#info] == 'bag_all' and input == true then
		options.args.c.args.filter_bags.args.bag_all.name = "None"
		CleanMyBag:SetOption({'c', 'filter_bags', 'bag0'}, true)
		CleanMyBag:SetOption({'c', 'filter_bags', 'bag1'}, true)
		CleanMyBag:SetOption({'c', 'filter_bags', 'bag2'}, true)
		CleanMyBag:SetOption({'c', 'filter_bags', 'bag3'}, true)
		CleanMyBag:SetOption({'c', 'filter_bags', 'bag4'}, true)
	elseif info[#info] == 'bag_all' and input == false then
		options.args.c.args.filter_bags.args.bag_all.name = "All"
		CleanMyBag:SetOption({'c', 'filter_bags', 'bag0'}, false)
		CleanMyBag:SetOption({'c', 'filter_bags', 'bag1'}, false)
		CleanMyBag:SetOption({'c', 'filter_bags', 'bag2'}, false)
		CleanMyBag:SetOption({'c', 'filter_bags', 'bag3'}, false)
		CleanMyBag:SetOption({'c', 'filter_bags', 'bag4'}, false)
	end
	if info[#info] == 'bag0' and input == false then
		options.args.c.args.filter_bags.args.bag_all.name = "All"
		CleanMyBag.db.char.bag_all = false
	end
	if info[#info] == 'bag1' and input == false then
		options.args.c.args.filter_bags.args.bag_all.name = "All"
		CleanMyBag.db.char.bag_all = false
	end
	if info[#info] == 'bag2' and input == false then
		options.args.c.args.filter_bags.args.bag_all.name = "All"
		CleanMyBag.db.char.bag_all = false
	end
	if info[#info] == 'bag3' and input == false then
		options.args.c.args.filter_bags.args.bag_all.name = "All"
		CleanMyBag.db.char.bag_all = false
	end
	if info[#info] == 'bag4' and input == false then
		options.args.c.args.filter_bags.args.bag_all.name = "All"
		CleanMyBag.db.char.bag_all = false
	end
	
	if CleanMyBag.db.char.bag0 and CleanMyBag.db.char.bag1 and CleanMyBag.db.char.bag2 and CleanMyBag.db.char.bag3 and CleanMyBag.db.char.bag4 then
		options.args.c.args.filter_bags.args.bag_all.name = "None"
		CleanMyBag.db.char.bag_all = true
	end
end

function CleanMyBag:GetFilters(info)
	return CleanMyBag.db.char.cfilters
end

function CleanMyBag:SetFilters(info, input)
	filters = {}
	CleanMyBag.db.char.cfilters = ""
	if input ~= '' then
		local new_filter = {}
		local fstring = split_str(input, "[^\r\n]+", false);
			for i,f in ipairs(fstring) do
				new_filter = {}
				for w in f:gmatch("[%a]+%s[(=)(!=)(<)]+%s\"[%a%d%s&]+[\"$]") do
					if CleanMyBag.db.char.dev_debug then
						print_str(': matched filter found: ' .. w)
					end
					table.insert(new_filter, w)
				end
				if table.getn(new_filter) > 0 then
					table.insert(filters, new_filter)
					CleanMyBag.db.char.cfilters = CleanMyBag.db.char.cfilters .. table.concat(new_filter, " and ") .. "\r"
				end 
			end
			if CleanMyBag.db.char.dev_debug then
				print_str('filters found: ' .. table.getn(filters))
			end
			if table.getn( filters ) > 0 then
				for mi,mf in pairs(filters) do
					if CleanMyBag.db.char.dev_debug then
					end
					for index,fval in pairs(mf) do
						if CleanMyBag.db.char.dev_debug then
							print_str(' breaking down filter')
						end
						local pattern = string.explode(fval:gsub("([%a]+)%s([(=)(!=)(<)]+)%s(\"[%a%d%s&]+\")", "%1_%2_%3"), '_')
						if CleanMyBag.db.char.dev_debug then
							print_str(' pattern length: ' .. table.getn(pattern))
						end
						if table.getn(pattern) == 3 then
							pattern[3] = trim_str(pattern[3], "\"")
							if pattern[1] == "name" then
								filters[mi] = {}
								filters[mi][1] = pattern
								break
							else
								filters[mi][index] = pattern
							end
						else
							if CleanMyBag.db.char.dev_debug then
								print_str(' rule isnt correct')
							end
						end
					end
				end
			end
	else
		if CleanMyBag.db.char.dev_debug then
			print_str(': input == nil')
		end
	end
end

function CleanMyBag:GetSafeList(info)
	return CleanMyBag.db.char.safelist
end

function CleanMyBag:SetSafeList(info, input)
	safe = {}
	CleanMyBag.db.char.safelist = ""
	if input ~= '' then
		local safeItems = string.explode(input, ",")
		for i,f in ipairs(safeItems) do
			if f ~= '' then
				table.insert(safe, f)
			end
		end
		CleanMyBag.db.char.safelist = table.concat(safe, ", ")
	end
end

function CleanMyBag:CHAT_MSG_LOOT(event, msg)
	if GetNumSubgroupMembers() > 0 or GetNumGroupMembers() > 0 then
		 ProcessRollResult(event, msg)
	end
end

function ProcessRollResult(event, msg)
	local filtered = msg:gsub("(%a+)%s(%a+)%s(%a+):%s(.-)%.", "%1_%2_%3_%4")
	local wonMsg = string.explode(filtered, '_')
	if table.getn(wonMsg) == 4 then
		if wonMsg[4] ~= nil then
			local lootName = GetItemInfo(wonMsg[4])
			local itemIndex = item_index(groupLoot, lootName)
			if itemIndex ~= false then
				table.remove(groupLoot, itemIndex)
			end
		end
	end
end

function print_str(s)
	DEFAULT_CHAT_FRAME:AddMessage('Clean My Bag' .. s, 0.1, 0.7, 1.0);
end

function split_str(s, p, kv)
	t = {}
	for key,value in s:gmatch(p) do
		if kv == true then
			if value == nil then value = "nil" end
			t[key] = value
		else
			table.insert(t, key)
		end
	end
	if table.getn(t) == 0 then table.insert(t, s) end
	return t;
end

function string.explode(str, sep) 
	local pos, t = 1, {}
	if #sep == 0 or #str == 0 then return end
	for s, e in function() return string.find(str, sep, pos) end do
		table.insert(t, string.trim(string.sub(str, pos, s-1)))
		pos = e+1
	end
	table.insert(t, string.trim(string.sub(str, pos)))
	return t
end

function trim_str(s, tchar)
	return (s:gsub("^" .. tchar .. "*(.-)" .. tchar .. "*$", "%1"))
end

function in_loot_table( lootName )
	for i=1,table.getn(loot) do
		if loot[i][1] == lootName then
			return i
		end
	end
	return -1
end

function in_table(t, sv)
	for _,v in pairs(t) do
		if v == sv then
			return true
		end
	end
	return false
end

function item_index(haystack, needle)
	for i=1,table.getn(haystack) do
		if haystack[i] == needle then
			return i
		end
	end
	return false
end